//+------------------------------------------------------------------+
//|                                                  SOM_ex1-hex.mq5 |
//|                        Copyright 2011, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2011, MetaQuotes Software Corp."
#property link      "http://www.mql5.com"
#property version   "1.00"

//  ,      
input bool HexagonalCell=true;
//  ,    
input bool ShowBorders=true;

//---   cIntBMP
#include <cIntBMP.mqh>
//+------------------------------------------------------------------+
//|   SOM-                                              |
//+------------------------------------------------------------------+
class CSOMNode
  {
protected:
   int               m_x1;
   int               m_y1;
   int               m_x2;
   int               m_y2;

   double            m_x;
   double            m_y;
   double            m_weights[];
public:
   //---  
                     CSOMNode();
   //---  
                    ~CSOMNode();
   //---    
   void              InitNode(int x1,int y1,int x2,int y2);
   //---     
   double            X()  const { return(m_x);}
   double            Y()  const { return(m_y);}
   //---    
   void              GetCoordinates(int &x1,int &y1,int &x2,int &y2);
   //---     weight_index
   double            GetWeight(int weight_index);
   //---       
   double            CalculateDistance(double &vector[]);
   //---    
   void              AdjustWeights(double &vector[],double learning_rate,double influence);
  };
//+------------------------------------------------------------------+
//|                                                 |
//+------------------------------------------------------------------+
void CSOMNode::CSOMNode()
  {
   m_x1=0;
   m_y1=0;
   m_x2=0;
   m_y2=0;
   m_x=0.0;
   m_y=0.0;
  };
//+------------------------------------------------------------------+
//|                                       |
//+------------------------------------------------------------------+
void CSOMNode::InitNode(int x1,int y1,int x2,int y2)
  {
//---   
   m_x1=x1;
   m_y1=y1;
   m_x2=x2;
   m_y2=y2;
//---    
   m_x=x1+MathAbs((x2-x1))/2;
   m_y=y1+MathAbs((y2-y1))/2;
//---      
   ArrayResize(m_weights,3);
//---    ;
   for(int i=0; i<3; i++) {m_weights[i]=255.0*rand()/32768;}
  };
//+------------------------------------------------------------------+
//|                                                  |
//+------------------------------------------------------------------+
void CSOMNode::~CSOMNode()
  {
   ArrayResize(m_weights,0);
  };
//+------------------------------------------------------------------+
//|     weight_index                    |
//+------------------------------------------------------------------+
double CSOMNode::GetWeight(int weight_index)
  {
   if(weight_index>=0 && weight_index<3)
     {
      return(m_weights[weight_index]);
     }
   else return(0);
  }
//+------------------------------------------------------------------+
//|                                   |
//+------------------------------------------------------------------+
void CSOMNode::GetCoordinates(int &x1,int &y1,int &x2,int &y2)
  {
   x1=m_x1;
   y1=m_y1;
   x2=m_x2;
   y2=m_y2;
  }
//+------------------------------------------------------------------+
//|              |
//+------------------------------------------------------------------+
double CSOMNode::CalculateDistance(double &vector[])
  {
   double dist_sqr=0;
   for(int i=0; i<3; i++)
     {
      dist_sqr=dist_sqr+(vector[i]-m_weights[i])*(vector[i]-m_weights[i]);
     }
   return(dist_sqr);
  };
//+------------------------------------------------------------------+
//|                |
//+------------------------------------------------------------------+
void CSOMNode::AdjustWeights(double &vector[],double learning_rate,double influence)
  {
   for(int i=0; i<3; i++)
     {
      m_weights[i]=m_weights[i]+learning_rate*influence*(vector[i]-m_weights[i]);
     }
  }
//+------------------------------------------------------------------+
//|  SOM                                                        |
//+------------------------------------------------------------------+
class CSOM
  {
protected:
   //---     BMP-
   cIntBMP           m_bmp;
   //---      
   int               m_gridmode;
   //---  BMP-
   int               m_xsize;
   int               m_ysize;
   //---  
   int               m_xcells;
   int               m_ycells;
   //---    
   CSOMNode          m_som_nodes[];
   //---    
   int               m_total_training_sets;
   //---    
   double            m_training_sets_array[];
protected:
   double            m_map_radius;
   double            m_time_constant;

   double            m_initial_learning_rate;
   int               m_iterations;

public:
   //---  
                     CSOM();
   //---  
                    ~CSOM();
   //---   
   void              InitParameters(int iterations,int xcells,int ycells,int bmpwidth,int bmpheight);
   //---        
   int               BestMatchingNode(double &vector[]);
   //---   
   void              Train();
   //---    
   void              Render();
   //---   bmp-  
   void              ShowBMP(bool back);
   //---      
   void              AddVectorToTrainingSet(double &vector[]);
   //---       
   void              ShowPattern(double c1,double c2,double c3,string name);
   //---    
   void              AddTrainPattern(double c1,double c2,double c3);
   //---    col  RGB-
   void              ColToRGB(int col,int &r,int &g,int &b);
   //---     RGB-
   int               RGB256(int r,int g,int b) const {return(r+256*g+65536*b);      }
   //---   ,  BMP-  
   void              NetDeinit();
  };
//+------------------------------------------------------------------+
//|                                                       |
//+------------------------------------------------------------------+
void CSOM::CSOM()
  {
   m_iterations=10;
   m_initial_learning_rate=0.1;
   m_total_training_sets=0;
   ArrayResize(m_training_sets_array,0);
  }
//+------------------------------------------------------------------+
//|                                                        |
//+------------------------------------------------------------------+
void CSOM::~CSOM()
  {
   ArrayResize(m_training_sets_array,0);
  }
//+------------------------------------------------------------------+
//|                                         |
//+------------------------------------------------------------------+
void CSOM::InitParameters(int iterations,int xcells,int ycells,int bmpwidth,int bmpheight)
  {
   m_xsize=bmpwidth;
   m_ysize=bmpheight;

   m_xcells=xcells;
   m_ycells=ycells;

//---       
   m_gridmode=int(ChartGetInteger(0,CHART_SHOW_GRID,0));

//---      
   m_iterations=iterations;
   int cellwidth=m_xsize/m_xcells;
   int cellheight=m_ysize/m_ycells;

//---  BMP-
   if(!HexagonalCell) {m_bmp.Create(m_xsize,m_ysize,clrWhite);}
   else m_bmp.Create(m_xsize+cellwidth,m_ysize+cellheight/2,clrWhite);

//---        
   ArrayResize(m_som_nodes,xcells*ycells);
   int ind=0;
//---  
   for(int i=0; i<m_xcells; i++)
     {
      for(int j=0; j<m_ycells; j++)
        {
         m_som_nodes[ind].InitNode(i*cellwidth,j*cellheight,(i+1)*cellwidth,(j+1)*cellheight);
         ind++;
        }
     }
//---    
   if(m_xsize>m_ysize) {m_map_radius=m_xsize/2.0;}
   else {m_map_radius=m_ysize/2.0;}
//Print("m_map_radius=",m_map_radius);
   m_time_constant=1.0*m_iterations/MathLog(m_map_radius);
   Print("m_time_constant=",m_time_constant);
  }
//+------------------------------------------------------------------+
//|                              |
//+------------------------------------------------------------------+
int CSOM::BestMatchingNode(double &vector[])
  {
   int    min_ind=0;
   double min_dist=m_som_nodes[min_ind].CalculateDistance(vector);
   int total_nodes= ArraySize(m_som_nodes);
   for(int i=0; i<total_nodes; i++)
     {
      if(m_som_nodes[i].CalculateDistance(vector)<min_dist)
        {
         min_dist=m_som_nodes[i].CalculateDistance(vector);
         min_ind=i;
        };
     }
   return(min_ind);
  }
//+------------------------------------------------------------------+
//|                                              |
//+------------------------------------------------------------------+
void CSOM::Train()
  {
   double learning_rate=m_initial_learning_rate;
   int iter=0;
   double datavector[];
   ArrayResize(datavector,3);

   int total_nodes=ArraySize(m_som_nodes);

   int frame=0;
   do
     {
      //---        
      int ind=int(1.0*m_total_training_sets*rand()/32768);
      //---  datavector  
      datavector[0]=m_training_sets_array[3*(ind)];
      datavector[1]=m_training_sets_array[3*(ind)+1];
      datavector[2]=m_training_sets_array[3*(ind)+2];
      //---    ,    datavector
      int winningnode=BestMatchingNode(datavector);
      //---    
      double neighbourhood_radius=m_map_radius*MathExp(-1.0*iter/m_time_constant);

      //---     
      for(int i=0; i<total_nodes; i++)
        {
         //---        m_WinningNodeIndex
         double DistToNodeSqr=(m_som_nodes[winningnode].X()-m_som_nodes[i].X())*(m_som_nodes[winningnode].X()-m_som_nodes[i].X())
                              +(m_som_nodes[winningnode].Y()-m_som_nodes[i].Y())*(m_som_nodes[winningnode].Y()-m_som_nodes[i].Y());
         //---                                 
         double WS=neighbourhood_radius*neighbourhood_radius;
         //---    ,   
         if(DistToNodeSqr<WS)
           {
            //---     
            double influence=MathExp(-DistToNodeSqr/(2*WS));
            //---       
            m_som_nodes[i].AdjustWeights(datavector,learning_rate,influence);
           }
        }
      //---  ( iter)   
      learning_rate=m_initial_learning_rate*MathExp(-1.0*iter/m_iterations);
      //---   
      iter++;
      //---  ,    
      //---      N- 
/*if(iter%100==0)
        {
         frame++;
         Print(" ",iter,"  ",m_iterations," ",int(iter*100.0/m_iterations),"%");
         Render();
         ShowBMP(false);
        }*/
     }
   //---     ,       
   while(iter<m_iterations);

   Print(" .");
  }
//+------------------------------------------------------------------+
//|   bmp-                 |
//+------------------------------------------------------------------+
void CSOM::ShowBMP(bool back)
  {
//---      
   ChartSetInteger(0,CHART_SHOW_GRID,0);
//---  BMP-  
   m_bmp.Save("bmpimg",true);
//---       
   m_bmp.Show(0,0,"bmpimg","SOM");
//---    
   ObjectSetInteger(0,"SOM",OBJPROP_BACK,back);
  }
//+------------------------------------------------------------------+
//|                                |
//+------------------------------------------------------------------+
void CSOM::AddVectorToTrainingSet(double &vector[])
  {
   m_total_training_sets++;
   ArrayResize(m_training_sets_array,3*m_total_training_sets);
   for(int i=0; i<3; i++)
     {
      m_training_sets_array[3*(m_total_training_sets-1)+i]=vector[i];
     }
  }
//+------------------------------------------------------------------+
//|                       |
//+------------------------------------------------------------------+
void CSOM::ShowPattern(double c1,double c2,double c3,string name)
  {
   double datavector[3];
   datavector[0]=c1;
   datavector[1]=c2;
   datavector[2]=c3;
   int WinningNodeIndex=BestMatchingNode(datavector);
   int ind=BestMatchingNode(datavector);

   int x1,y1,x2,y2;
   m_som_nodes[ind].GetCoordinates(x1,y1,x2,y2);

   int r= int(m_som_nodes[ind].GetWeight(0));
   int g= int(m_som_nodes[ind].GetWeight(1));
   int b= int(m_som_nodes[ind].GetWeight(2));

   if(HexagonalCell)
     {
      //--- 
      int x_size=int (MathAbs(x2-x1));
      int y_size=int (MathAbs(y2-y1));
      y1=y1+y_size/4;
      y2=y2+y_size/4;
     }

   int xc=(x1+x2)/2;
   int yc=(y1+y2)/2;

   if(r*r+g*g+b*b<=128*128)
     {
      m_bmp.TypeText(xc+1,yc+1,name,clrBlack);
      m_bmp.TypeText(xc,yc,name,clrWhite);
     }
   else
     {
      m_bmp.TypeText(xc+1,yc+1,name,clrBlack);
      m_bmp.TypeText(xc,yc,name,clrWhite);
     }
   Print(" ",name,":     =",ind);
  }
//+------------------------------------------------------------------+
//|                        |
//+------------------------------------------------------------------+
void CSOM::AddTrainPattern(double c1,double c2,double c3)
  {
   double vector[3];
   vector[0]=c1;
   vector[1]=c2;
   vector[2]=c3;
   AddVectorToTrainingSet(vector);
  }
//+------------------------------------------------------------------+
//|    col  RGB-                    |
//+------------------------------------------------------------------+
void CSOM::ColToRGB(int col,int &r,int &g,int &b)
  {
   r=col & 0xFF;
   g=(col>>8)& 0xFF;
   b=(col>>16)& 0xFF;
  }
//+------------------------------------------------------------------+
//|      ( m_bmp)                 |
//+------------------------------------------------------------------+
void CSOM::Render()
  {
   int ind=0;
   for(int i=0; i<m_xcells; i++)
     {
      for(int j=0; j<m_ycells; j++)
        {
         int x1,y1,x2,y2;
         m_som_nodes[ind].GetCoordinates(x1,y1,x2,y2);
         int r = int(m_som_nodes[ind].GetWeight(0));
         int g = int(m_som_nodes[ind].GetWeight(1));
         int b = int(m_som_nodes[ind].GetWeight(2));
         int col=RGB256(r,g,b);

         if(!HexagonalCell)
           {
            //---  
            m_bmp.DrawRectangle(x1,y1,x2,y2,col,true);
            if(ShowBorders) m_bmp.DrawRectangle(x1,y1,x2,y2,clrBlack,false);
           }
         else
           {
            //--- 
            int x_size=int (MathAbs(x2-x1));
            int y_size=int (MathAbs(y2-y1));
            if(j%2==0)
              {
               x1=x1+x_size/2;
               x2=x2+x_size/2;
              }
            y1=y1+y_size/4;
            y2=y2+y_size/4;

            m_bmp.DrawRectangle(x1,y1+y_size/4,x2,y2-y_size/4,col,true);
            m_bmp.DrawTriangle(x1,y1+y_size/4,x1+x_size/2,y1-y_size/4,x2,y1+y_size/4,col,true);
            m_bmp.DrawTriangle(x1,y2-y_size/4,x2,y2-y_size/4,x1+x_size/2,y2+y_size/4,col,true);

            if(ShowBorders)
              {
               //---  
               int bcol=clrBlack;
               m_bmp.DrawLine(x1,y1+y_size/4,x1+x_size/2,y1-y_size/4,bcol);
               m_bmp.DrawLine(x1+x_size/2,y1-y_size/4,x2,y1+y_size/4,bcol);

               m_bmp.DrawLine(x1,y2-y_size/4,x1+x_size/2,y2+y_size/4,bcol);
               m_bmp.DrawLine(x2,y2-y_size/4,x2-x_size/2,y2+y_size/4,bcol);

               m_bmp.DrawLine(x1,y1+y_size/4,x1,y2-y_size/4,bcol);
               m_bmp.DrawLine(x2,y1+y_size/4,x2,y2-y_size/4,bcol);
              }

           }
         ind++;
        }
     }
   return;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM::NetDeinit()
  {
//---      
   ChartSetInteger(0,CHART_SHOW_GRID,m_gridmode);
//---   SOM  
   ObjectDelete(0,"SOM");
  }
//+------------------------------------------------------------------+
//|  CSOMWeb                                                    |
//+------------------------------------------------------------------+
class CSOMWeb : public CSOM
  {
public:
//---    col    ( ,  ShowPattern)
   void              AddTrainColor(int col);
//---         ( ,  ShowPattern)
   void              ShowColor(int col,string name);
  };
//+------------------------------------------------------------------+
//|    col                                       |
//|    ( ,  ShowPattern)           |
//+------------------------------------------------------------------+
void CSOMWeb::AddTrainColor(int col)
  {
   double vector[];
   ArrayResize(vector,3);
   int r=0;
   int g=0;
   int b=0;
   ColToRGB(col,r,g,b);
   vector[0]=r;
   vector[1]=g;
   vector[2]=b;
   AddVectorToTrainingSet(vector);
   ArrayResize(vector,0);
  }
//+------------------------------------------------------------------+
//|                                 |
//|    ( ,  ShowPattern)    |
//+------------------------------------------------------------------+
void CSOMWeb::ShowColor(int col,string name)
  {
   int r=0;
   int g=0;
   int b=0;
   ColToRGB(col,r,g,b);
   ShowPattern(r,g,b,name);
  }

//--- web colors array
color web_colors[132]=
  {
   clrBlack,clrDarkGreen,clrDarkSlateGray,clrOlive,clrGreen,
   clrTeal,clrNavy,clrPurple,clrMaroon,clrIndigo,clrMidnightBlue,
   clrDarkBlue,clrDarkOliveGreen,clrSaddleBrown,clrForestGreen,clrOliveDrab,
   clrSeaGreen,clrDarkGoldenrod,clrDarkSlateBlue,clrSienna,clrMediumBlue,
   clrBrown,clrDarkTurquoise,clrDimGray,clrLightSeaGreen,clrDarkViolet,
   clrFireBrick,clrMediumVioletRed,clrMediumSeaGreen,clrChocolate,clrCrimson,
   clrSteelBlue,clrGoldenrod,clrMediumSpringGreen,clrLawnGreen,clrCadetBlue,
   clrDarkOrchid,clrYellowGreen,clrLimeGreen,clrOrangeRed,clrDarkOrange,
   clrOrange,clrGold,clrYellow,clrChartreuse,clrLime,clrSpringGreen,
   clrAqua,clrDeepSkyBlue,clrBlue,clrMagenta,clrRed,clrGray,
   clrSlateGray,clrPeru,clrBlueViolet,clrLightSlateGray,clrDeepPink,
   clrMediumTurquoise,clrDodgerBlue,clrTurquoise,clrRoyalBlue,clrSlateBlue,
   clrDarkKhaki,clrIndianRed,clrMediumOrchid,clrGreenYellow,clrMediumAquamarine,
   clrDarkSeaGreen,clrTomato,clrRosyBrown,clrOrchid,clrMediumPurple,
   clrPaleVioletRed,clrCoral,clrCornflowerBlue,clrDarkGray,clrSandyBrown,
   clrMediumSlateBlue,clrTan,clrDarkSalmon,clrBurlyWood,clrHotPink,
   clrSalmon,clrViolet,clrLightCoral,clrSkyBlue,clrLightSalmon,
   clrPlum,clrKhaki,clrLightGreen,clrAquamarine,clrSilver,
   clrLightSkyBlue,clrLightSteelBlue,clrLightBlue,clrPaleGreen,clrThistle,
   clrPowderBlue,clrPaleGoldenrod,clrPaleTurquoise,clrLightGray,clrWheat,
   clrNavajoWhite,clrMoccasin,clrLightPink,clrGainsboro,clrPeachPuff,
   clrPink,clrBisque,clrLightGoldenrod,clrBlanchedAlmond,clrLemonChiffon,
   clrBeige,clrAntiqueWhite,clrPapayaWhip,clrCornsilk,clrLightYellow,
   clrLightCyan,clrLinen,clrLavender,clrMistyRose,clrOldLace,
   clrWhiteSmoke,clrSeashell,clrIvory,clrHoneydew,clrAliceBlue,
   clrLavenderBlush,clrMintCream,clrSnow,clrWhite
  };

CSOMWeb KohonenMap;
//+------------------------------------------------------------------+
//| Expert initialization function                                   |
//+------------------------------------------------------------------+
void OnInit()
  {
   MathSrand(200);
   int total_web_colors=ArraySize(web_colors);

//---  ,  10000 
//---  50x50 ,   500x500
   KohonenMap.InitParameters(10000,50,50,500,500);
//--  RGB-      

//--   web-   
   for(int i=0; i<total_web_colors; i++)
     {
      KohonenMap.AddTrainColor(web_colors[i]);
     }
//---  
   KohonenMap.Train();
//---    
   KohonenMap.Render();

//---          
   for(int i=0; i<total_web_colors; i++)
     {
      KohonenMap.ShowColor(web_colors[i],ColorToString(web_colors[i],true));
     }
//---    
   KohonenMap.ShowBMP(false);
  }
//+------------------------------------------------------------------+
//| Expert deinitialization function                                 |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
   KohonenMap.NetDeinit();
  }
//+------------------------------------------------------------------+
//| Expert tick function                                             |
//+------------------------------------------------------------------+
void OnTick()
  {
//---
  }
//+------------------------------------------------------------------+
