//+------------------------------------------------------------------+
//|                                                      SOM_ex5.mq5 |
//|                        Copyright 2011, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2011, MetaQuotes Software Corp."
#property link      "http://www.mql5.com"
#property version   "1.00"

//  ,      
input bool HexagonalCell=true;
//  ,    
input bool ShowBorders=false;

//---   cIntBMP
#include <cIntBMP.mqh>
//+------------------------------------------------------------------+
//|   SOM-                                              |
//+------------------------------------------------------------------+
class CSOMNode
  {
protected:
   int               m_x1;
   int               m_y1;
   int               m_x2;
   int               m_y2;

   double            m_x;
   double            m_y;
   double            m_weights[];
public:
   //---  
                     CSOMNode();
   //---  
                    ~CSOMNode();
   //---    
   void              InitNode(int x1,int y1,int x2,int y2);
   //---     
   double            X()  const { return(m_x);}
   double            Y()  const { return(m_y);}
   //---    
   void              GetCoordinates(int &x1,int &y1,int &x2,int &y2);
   //---     weight_index
   double            GetWeight(int weight_index);
   //---       
   double            CalculateDistance(double &vector[]);
   //---    
   void              AdjustWeights(double &vector[],double learning_rate,double influence);
  };
//+------------------------------------------------------------------+
//|                                                 |
//+------------------------------------------------------------------+
void CSOMNode::CSOMNode()
  {
   m_x1=0;
   m_y1=0;
   m_x2=0;
   m_y2=0;
   m_x=0.0;
   m_y=0.0;
  };
//+------------------------------------------------------------------+
//|                                       |
//+------------------------------------------------------------------+
void CSOMNode::InitNode(int x1,int y1,int x2,int y2)
  {
//---   
   m_x1=x1;
   m_y1=y1;
   m_x2=x2;
   m_y2=y2;
//---    
   m_x=x1+MathAbs((x2-x1))/2;
   m_y=y1+MathAbs((y2-y1))/2;
//---      
   ArrayResize(m_weights,4);
//---    ;
   for(int i=0; i<4; i++) {m_weights[i]=1.0*rand()/32768;}
  };
//+------------------------------------------------------------------+
//|                                                  |
//+------------------------------------------------------------------+
void CSOMNode::~CSOMNode()
  {
   ArrayResize(m_weights,0);
  };
//+------------------------------------------------------------------+
//|     weight_index                    |
//+------------------------------------------------------------------+
double CSOMNode::GetWeight(int weight_index)
  {
   if(weight_index>=0 && weight_index<4)
     {
      return(m_weights[weight_index]);
     }
   else return(0);
  }
//+------------------------------------------------------------------+
//|                                   |
//+------------------------------------------------------------------+
void CSOMNode::GetCoordinates(int &x1,int &y1,int &x2,int &y2)
  {
   x1=m_x1;
   y1=m_y1;
   x2=m_x2;
   y2=m_y2;
  }
//+------------------------------------------------------------------+
//|              |
//+------------------------------------------------------------------+
double CSOMNode::CalculateDistance(double &vector[])
  {
   double dist_sqr=0;
   for(int i=0; i<4; i++)
     {
      dist_sqr=dist_sqr+(vector[i]-m_weights[i])*(vector[i]-m_weights[i]);
     }
   return(dist_sqr);
  };
//+------------------------------------------------------------------+
//|                |
//+------------------------------------------------------------------+
void CSOMNode::AdjustWeights(double &vector[],double learning_rate,double influence)
  {
   for(int i=0; i<4; i++)
     {
      m_weights[i]=m_weights[i]+learning_rate*influence*(vector[i]-m_weights[i]);
     }
  }
//+------------------------------------------------------------------+
//|  SOM                                                        |
//+------------------------------------------------------------------+
class CSOM
  {
protected:
   uchar             Palette[768];
   //---      BMP-
   cIntBMP           m_bmp[];
   //---      
   int               m_gridmode;
   //---  BMP-
   int               m_xsize;
   int               m_ysize;
   //---  
   int               m_xcells;
   int               m_ycells;
   //---    
   CSOMNode          m_som_nodes[];
   //---    
   int               m_total_training_sets;
   //---    
   double            m_training_sets_array[];
   double            m_max_values[];
   double            m_min_values[];

protected:
   double            m_map_radius;
   double            m_time_constant;

   double            m_initial_learning_rate;
   int               m_iterations;

public:
   //---  
                     CSOM();
   //---  
                    ~CSOM();
   //---   
   void              InitParameters(int iterations,int xcells,int ycells,int bmpwidth,int bmpheight);
   //---        
   int               BestMatchingNode(double &vector[]);
   //---   
   void              Train();
   //---    
   void              Render();
   //---   bmp-  
   void              ShowBMP(bool back);
   //---      
   void              AddVectorToTrainingSet(double &vector[]);
   //---       
   void              ShowPattern(double c1,double c2,double c3,double c4,string name);
   //---    
   void              AddTrainPattern(double c1,double c2,double c3,double c4);
   //---    col  RGB-
   void              ColToRGB(int col,int &r,int &g,int &b);
   //---     RGB-
   int               RGB256(int r,int g,int b) const {return(r+256*g+65536*b);      }
   //---     CMYK- 
   int               CMYK2Col(uchar c,uchar m,uchar y,uchar k);
   //---       
   int               GetPalColor(uchar index);
   //---    
   void              Blend(uchar c1,uchar c2,uchar r1,uchar g1,uchar b1,uchar r2,uchar g2,uchar b2);
   //---   
   void              RenderCell(int img,int col,int node_index,bool cell_even);
   //---   ,  BMP-  
   void              NetDeinit();
  };
//+------------------------------------------------------------------+
//|                                                       |
//+------------------------------------------------------------------+
void CSOM::CSOM()
  {
   m_iterations=10;
   m_initial_learning_rate=0.1;
   m_total_training_sets=0;
   ArrayResize(m_training_sets_array,0);
  }
//+------------------------------------------------------------------+
//|                                                        |
//+------------------------------------------------------------------+
void CSOM::~CSOM()
  {
   ArrayResize(m_training_sets_array,0);
  }
//+------------------------------------------------------------------+
//|                                         |
//+------------------------------------------------------------------+
void CSOM::InitParameters(int iterations,int xcells,int ycells,int bmpwidth,int bmpheight)
  {
   m_xsize=bmpwidth;
   m_ysize=bmpheight;

   m_xcells=xcells;
   m_ycells=ycells;

//---       
   m_gridmode=int(ChartGetInteger(0,CHART_SHOW_GRID,0));

//---      
   m_iterations=iterations;
   int cellwidth=m_xsize/m_xcells;
   int cellheight=m_ysize/m_ycells;

//---  BMP-
   ArrayResize(m_bmp,4+1);
   for(int i=0; i<4+1; i++)
     {
      //m_bmp[i].Create(m_xsize,m_ysize+15,clrWhite);
      if(!HexagonalCell) {m_bmp[i].Create(m_xsize,m_ysize,clrWhite);}
      else m_bmp[i].Create(m_xsize+cellwidth,m_ysize+cellheight/2,clrWhite);
     }

   Blend(0,255,0,0,0,255,0,0);

   Blend(0,1*64-1,98,98,255,98,255,255);
   Blend(1*64,2*64-1,98,255,255,98,255,98);
   Blend(2*64,3*64-1,98,255,98,255,255,98);
   Blend(3*64,4*64-1,255,255,98,255,98,98);

//---        
   ArrayResize(m_som_nodes,xcells*ycells);
   int ind=0;
//---  
   for(int i=0; i<m_xcells; i++)
     {
      for(int j=0; j<m_ycells; j++)
        {
         m_som_nodes[ind].InitNode(i*cellwidth,j*cellheight,(i+1)*cellwidth,(j+1)*cellheight);
         ind++;
        }
     }
//---    
   if(m_xsize>m_ysize) {m_map_radius=m_xsize/2.0;}
   else {m_map_radius=m_ysize/2.0;}
//Print("m_map_radius=",m_map_radius);
   m_time_constant=1.0*m_iterations/MathLog(m_map_radius);
   Print("m_time_constant=",m_time_constant);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM::Blend(uchar c1,uchar c2,uchar r1,uchar g1,uchar b1,uchar r2,uchar g2,uchar b2)
  {
   int n=(c2-c1);
   for(int i=0; i<=n; i++)
     {
      if((c1+i+2)<ArraySize(Palette))
        {
         Palette[3*(c1+i)]=uchar(MathRound(1*(r1*(n-i)+r2*i)*1.0/n));
         Palette[3*(c1+i)+1]=uchar (MathRound(1*(g1*(n-i) + g2*i) * 1.0/ n));
         Palette[3*(c1+i)+2]=uchar (MathRound(1*(b1*(n-i) + b2*i) * 1.0/ n));
        }
     }
  }
//+------------------------------------------------------------------+
//|                          |
//+------------------------------------------------------------------+
int CSOM::GetPalColor(uchar index)
  {
   int ind=index;
   if(ind<=0) {ind=0;}
   uchar r=Palette[3*(ind)];
   uchar g=Palette[3*(ind)+1];
   uchar b=Palette[3*(ind)+2];
   return(RGB256(r,g,b));
  }
//+------------------------------------------------------------------+
//|                              |
//+------------------------------------------------------------------+
int CSOM::BestMatchingNode(double &vector[])
  {
   int    min_ind=0;
   double min_dist=m_som_nodes[min_ind].CalculateDistance(vector);
   int total_nodes= ArraySize(m_som_nodes);
   for(int i=0; i<total_nodes; i++)
     {
      if(m_som_nodes[i].CalculateDistance(vector)<min_dist)
        {
         min_dist=m_som_nodes[i].CalculateDistance(vector);
         min_ind=i;
        };
     }
   return(min_ind);
  }
//+------------------------------------------------------------------+
//|                                              |
//+------------------------------------------------------------------+
void CSOM::Train()
  {
   double learning_rate=m_initial_learning_rate;
   int iter=0;
   double datavector[];
   ArrayResize(datavector,4);

//---        
   ArrayResize(m_max_values,4);
   ArrayResize(m_min_values,4);

   for(int j=0; j<4; j++)
     {
      double maxv=m_training_sets_array[4+j];
      double minv=m_training_sets_array[4+j];
      for(int i=1; i<m_total_training_sets; i++)
        {
         double v=m_training_sets_array[4*i+j];
         if(v>maxv) {maxv=v;}
         if(v<minv) {minv=v;}
        }
      m_max_values[j]=maxv;
      m_min_values[j]=minv;
      Print(j,"m_min_value=",m_min_values[j],"m_max_value=",m_max_values[j]);
     }

   int total_nodes=ArraySize(m_som_nodes);

   int frame=0;
   do
     {
      //---        
      int ind=int(1.0*m_total_training_sets*rand()/32768);
      //---  datavector  
      datavector[0]=m_training_sets_array[4*(ind)];
      datavector[1]=m_training_sets_array[4*(ind)+1];
      datavector[2]=m_training_sets_array[4*(ind)+2];
      datavector[3]=m_training_sets_array[4*(ind)+3];
      //---    ,    datavector
      int winningnode=BestMatchingNode(datavector);
      //---    
      double neighbourhood_radius=m_map_radius*MathExp(-1.0*iter/m_time_constant);

      //---     
      for(int i=0; i<total_nodes; i++)
        {
         //---        m_WinningNodeIndex
         double DistToNodeSqr=(m_som_nodes[winningnode].X()-m_som_nodes[i].X())*(m_som_nodes[winningnode].X()-m_som_nodes[i].X())
                              +(m_som_nodes[winningnode].Y()-m_som_nodes[i].Y())*(m_som_nodes[winningnode].Y()-m_som_nodes[i].Y());
         //---                                 
         double WS=neighbourhood_radius*neighbourhood_radius;
         //---    ,   
         if(DistToNodeSqr<WS)
           {
            //---     
            double influence=MathExp(-DistToNodeSqr/(2*WS));
            //---       
            m_som_nodes[i].AdjustWeights(datavector,learning_rate,influence);
           }
        }
      //---  ( iter)   
      learning_rate=m_initial_learning_rate*MathExp(-1.0*iter/m_iterations);
      //---   
      iter++;
      //---  ,    
      //---      N- 
      if(iter%100==0)
        {
         frame++;
         Print(" ",iter,"  ",m_iterations," ",int(iter*100.0/m_iterations),"%");
         Render();
         ShowBMP(false);
        }
     }
   //---     ,       
   while(iter<m_iterations);

   Print(" .");
  }//+------------------------------------------------------------------+
//|   bmp-                 |
//+------------------------------------------------------------------+
void CSOM::ShowBMP(bool back)
  {
   for(int i=0; i<4+1; i++)
     {
      //---  BMP-  
      m_bmp[i].Save("bmpimg"+IntegerToString(i),true);
      //---       
      m_bmp[i].Show(10+i*(15+m_xsize),10+0,"bmpimg"+IntegerToString(i),"SOM"+IntegerToString(i));
      //---    
      ObjectSetInteger(0,"SOM"+IntegerToString(i),OBJPROP_BACK,back);
     }
  }
//+------------------------------------------------------------------+
//|                                |
//+------------------------------------------------------------------+
void CSOM::AddVectorToTrainingSet(double &vector[])
  {
   m_total_training_sets++;
   ArrayResize(m_training_sets_array,4*m_total_training_sets);
   for(int i=0; i<4; i++)
     {
      m_training_sets_array[4*(m_total_training_sets-1)+i]=vector[i];
     }
  }
//+-----------------------------------------------------------------------+
//| Portions from Xcmyk                                                   |
//| http://www.codeproject.com/KB/applications/xcmyk.aspx                 |
//| Original:                                                             |
//| http://www.neuro.sfc.keio.ac.jp/~aly/polygon/info/color-space-faq.html|
//+-----------------------------------------------------------------------+
//  CMYK --> RGB
//  --------------------------------------------
//  Red   = 1-minimum(1,Cyan*(1-Black)+Black)
//  Green = 1-minimum(1,Magenta*(1-Black)+Black)
//  Blue  = 1-minimum(1,Yellow*(1-Black)+Black)
//+-----------------------------------------------------------------------+
int CSOM::CMYK2Col(uchar c,uchar m,uchar y,uchar k)
  {
   uchar r,g,b;
   double C,M,Y,K;
   C = c/255.0;
   M = m/255.0;
   Y = y/255.0;
   K = k/255.0;
   r = uchar (255*(1.0-MathMin(1.0,C*(1-K)+K)));
   g = uchar (255*(1.0-MathMin(1.0,M*(1-K)+K)));
   b = uchar (255*(1.0-MathMin(1.0,Y*(1-K)+K)));
   return(RGB256(r,g,b));
  }
//+------------------------------------------------------------------+
//|                       |
//+------------------------------------------------------------------+
void CSOM::ShowPattern(double c1,double c2,double c3,double c4,string name)
  {
   double datavector[4];
   datavector[0]=c1;
   datavector[1]=c2;
   datavector[2]=c3;
   datavector[3]=c4;
   int WinningNodeIndex=BestMatchingNode(datavector);
   int ind=BestMatchingNode(datavector);

   int x1,y1,x2,y2;
   m_som_nodes[ind].GetCoordinates(x1,y1,x2,y2);

   int c=int ((255*(m_som_nodes[ind].GetWeight(0)-m_min_values[0])/(m_max_values[0]-m_min_values[0])));
   int m=int ((255*(m_som_nodes[ind].GetWeight(1)-m_min_values[1])/(m_max_values[1]-m_min_values[1])));
   int y=int ((255*(m_som_nodes[ind].GetWeight(2)-m_min_values[2])/(m_max_values[2]-m_min_values[2])));
   int k=int ((255*(m_som_nodes[ind].GetWeight(3)-m_min_values[3])/(m_max_values[3]-m_min_values[3])));

   if(HexagonalCell)
     {
      //--- 
      int x_size=int (MathAbs(x2-x1));
      int y_size=int (MathAbs(y2-y1));
      y1=y1+y_size/4;
      y2=y2+y_size/4;
     }

   int xc=(x1+x2)/2;
   int yc=(y1+y2)/2;

   if(c+m+y+k<=4*128)
     {
      m_bmp[4].TypeText(xc+1,yc+1,name,clrBlack);
      m_bmp[4].TypeText(xc,yc,name,clrWhite);
     }
   else
     {
      m_bmp[4].TypeText(xc+1,yc+1,name,clrBlack);
      m_bmp[4].TypeText(xc,yc,name,clrWhite);
     }
   Print(" ",name,":     =",ind);
  }
//+------------------------------------------------------------------+
//|                        |
//+------------------------------------------------------------------+
void CSOM::AddTrainPattern(double c1,double c2,double c3,double c4)
  {
   double vector[4];
   vector[0]=c1;
   vector[1]=c2;
   vector[2]=c3;
   vector[3]=c4;
   AddVectorToTrainingSet(vector);
  }
//+------------------------------------------------------------------+
//|    col  RGB-                    |
//+------------------------------------------------------------------+
void CSOM::ColToRGB(int col,int &r,int &g,int &b)
  {
   r=col & 0xFF;
   g=(col>>8)& 0xFF;
   b=(col>>16)& 0xFF;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM::RenderCell(int img,int col,int node_index,bool cell_even)
  {
   int x1,y1,x2,y2;
   m_som_nodes[node_index].GetCoordinates(x1,y1,x2,y2);

   if(!HexagonalCell)
     {
      //---  
      m_bmp[img].DrawRectangle(x1,y1,x2,y2,col,true);
      if(ShowBorders) m_bmp[img].DrawRectangle(x1,y1,x2,y2,clrBlack,false);
     }
   else
     {
      //--- 
      int x_size=int (MathAbs(x2-x1));
      int y_size=int (MathAbs(y2-y1));
      if(cell_even)
        {
         x1=x1+x_size/2;
         x2=x2+x_size/2;
        }
      y1=15+y1+y_size/4;
      y2=15+y2+y_size/4;

      m_bmp[img].DrawRectangle(x1,y1+y_size/4,x2,y2-y_size/4,col,true);
      m_bmp[img].DrawTriangle(x1,y1+y_size/4,x1+x_size/2,y1-y_size/4,x2,y1+y_size/4,col,true);
      m_bmp[img].DrawTriangle(x1,y2-y_size/4,x2,y2-y_size/4,x1+x_size/2,y2+y_size/4,col,true);

      if(ShowBorders)
        {
         //---  
         int bcol=clrBlack;
         m_bmp[img].DrawLine(x1,y1+y_size/4,x1+x_size/2,y1-y_size/4,bcol);
         m_bmp[img].DrawLine(x1+x_size/2,y1-y_size/4,x2,y1+y_size/4,bcol);

         m_bmp[img].DrawLine(x1,y2-y_size/4,x1+x_size/2,y2+y_size/4,bcol);
         m_bmp[img].DrawLine(x2,y2-y_size/4,x2-x_size/2,y2+y_size/4,bcol);

         m_bmp[img].DrawLine(x1,y1+y_size/4,x1,y2-y_size/4,bcol);
         m_bmp[img].DrawLine(x2,y1+y_size/4,x2,y2-y_size/4,bcol);
        }
     }
  }
//+------------------------------------------------------------------+
//|      ( m_bmp)                 |
//+------------------------------------------------------------------+
void CSOM::Render()
  {
   int ind=0;
   for(int i=0; i<m_xcells; i++)
     {
      for(int j=0; j<m_ycells; j++)
        {
        //---        
         for(int k=0; k<4; k++)
           {
            //m_som_nodes[ind].GetCoordinates(x1,y1,x2,y2);
            double range=(m_max_values[k]-m_min_values[k]);
            char avg= char(255*(m_som_nodes[ind].GetWeight(k)-m_min_values[k])/range);
            int col=int(GetPalColor(avg));
            RenderCell(k,col,ind,(j%2==0));
           }
         //---       CMYK-
         char c=char ((255*(m_som_nodes[ind].GetWeight(0)-m_min_values[0])/(m_max_values[0]-m_min_values[0])));
         char m=char ((255*(m_som_nodes[ind].GetWeight(1)-m_min_values[1])/(m_max_values[1]-m_min_values[1])));
         char y=char ((255*(m_som_nodes[ind].GetWeight(2)-m_min_values[2])/(m_max_values[2]-m_min_values[2])));
         char k=char ((255*(m_som_nodes[ind].GetWeight(3)-m_min_values[3])/(m_max_values[3]-m_min_values[3])));
         int col=CMYK2Col(c,m,y,k);
         RenderCell(4,col,ind,(j%2==0));
         ind++;
        }
      //---        
      for(int m=0; m<4; m++)
        {
         for(int k=0; k<15; k++)
           {
            for(int l=0; l<m_xsize; l++)
              {
               uchar colind=uchar(MathRound(l*255.0/m_xsize));
               int col=int(GetPalColor(colind));
               m_bmp[m].DrawDot(l,k,col);
              }
           }
         //---         
         m_bmp[m].TypeText(0,0,DoubleToString(m_min_values[m],2),clrBlack);
         m_bmp[m].TypeText(m_xsize-25,0,DoubleToString(m_max_values[m],2),clrBlack);
        }
     }
   return;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM::NetDeinit()
  {
//---      
   ChartSetInteger(0,CHART_SHOW_GRID,m_gridmode);
//---   SOM  
   ObjectDelete(0,"SOM");
  }

CSOM KohonenMap;
//+------------------------------------------------------------------+
//| Expert initialization function                                   |
//+------------------------------------------------------------------+
void OnInit()
  {
   MathSrand(200);

//---  ,  10000 
//---  50x50 ,   300x300
   KohonenMap.InitParameters(10000,50,50,300,300);

//---      
// 
// , , , , 
   KohonenMap.AddTrainPattern(5.1,3.5,1.4,0.2); // setosa
   KohonenMap.AddTrainPattern(4.9,3.0,1.4,0.2); // setosa
   KohonenMap.AddTrainPattern(4.7,3.2,1.3,0.2); // setosa
   KohonenMap.AddTrainPattern(4.6,3.1,1.5,0.2); // setosa
   KohonenMap.AddTrainPattern(5.0,3.6,1.4,0.2); // setosa
   KohonenMap.AddTrainPattern(5.4,3.9,1.7,0.4); // setosa
   KohonenMap.AddTrainPattern(4.6,3.4,1.4,0.3); // setosa
   KohonenMap.AddTrainPattern(5.0,3.4,1.5,0.2); // setosa
   KohonenMap.AddTrainPattern(4.4,2.9,1.4,0.2); // setosa
   KohonenMap.AddTrainPattern(4.9,3.1,1.5,0.1); // setosa
   KohonenMap.AddTrainPattern(5.4,3.7,1.5,0.2); // setosa
   KohonenMap.AddTrainPattern(4.8,3.4,1.6,0.2); // setosa
   KohonenMap.AddTrainPattern(4.8,3.0,1.4,0.1); // setosa
   KohonenMap.AddTrainPattern(4.3,3.0,1.1,0.1); // setosa
   KohonenMap.AddTrainPattern(5.8,4.0,1.2,0.2); // setosa
   KohonenMap.AddTrainPattern(5.7,4.4,1.5,0.4); // setosa
   KohonenMap.AddTrainPattern(5.4,3.9,1.3,0.4); // setosa
   KohonenMap.AddTrainPattern(5.1,3.5,1.4,0.3); // setosa
   KohonenMap.AddTrainPattern(5.7,3.8,1.7,0.3); // setosa
   KohonenMap.AddTrainPattern(5.1,3.8,1.5,0.3); // setosa
   KohonenMap.AddTrainPattern(5.4,3.4,1.7,0.2); // setosa
   KohonenMap.AddTrainPattern(5.1,3.7,1.5,0.4); // setosa
   KohonenMap.AddTrainPattern(4.6,3.6,1.0,0.2); // setosa
   KohonenMap.AddTrainPattern(5.1,3.3,1.7,0.5); // setosa
   KohonenMap.AddTrainPattern(4.8,3.4,1.9,0.2); // setosa
   KohonenMap.AddTrainPattern(5.0,3.0,1.6,0.2); // setosa
   KohonenMap.AddTrainPattern(5.0,3.4,1.6,0.4); // setosa
   KohonenMap.AddTrainPattern(5.2,3.5,1.5,0.2); // setosa
   KohonenMap.AddTrainPattern(5.2,3.4,1.4,0.2); // setosa
   KohonenMap.AddTrainPattern(4.7,3.2,1.6,0.2); // setosa
   KohonenMap.AddTrainPattern(4.8,3.1,1.6,0.2); // setosa
   KohonenMap.AddTrainPattern(5.4,3.4,1.5,0.4); // setosa
   KohonenMap.AddTrainPattern(5.2,4.1,1.5,0.1); // setosa
   KohonenMap.AddTrainPattern(5.5,4.2,1.4,0.2); // setosa
   KohonenMap.AddTrainPattern(4.9,3.1,1.5,0.2); // setosa
   KohonenMap.AddTrainPattern(5.0,3.2,1.2,0.2); // setosa
   KohonenMap.AddTrainPattern(5.5,3.5,1.3,0.2); // setosa
   KohonenMap.AddTrainPattern(4.9,3.6,1.4,0.1); // setosa
   KohonenMap.AddTrainPattern(4.4,3.0,1.3,0.2); // setosa
   KohonenMap.AddTrainPattern(5.1,3.4,1.5,0.2); // setosa
   KohonenMap.AddTrainPattern(5.0,3.5,1.3,0.3); // setosa
   KohonenMap.AddTrainPattern(4.5,2.3,1.3,0.3); // setosa
   KohonenMap.AddTrainPattern(4.4,3.2,1.3,0.2); // setosa
   KohonenMap.AddTrainPattern(5.0,3.5,1.6,0.6); // setosa
   KohonenMap.AddTrainPattern(5.1,3.8,1.9,0.4); // setosa
   KohonenMap.AddTrainPattern(4.8,3.0,1.4,0.3); // setosa
   KohonenMap.AddTrainPattern(5.1,3.8,1.6,0.2); // setosa
   KohonenMap.AddTrainPattern(4.6,3.2,1.4,0.2); // setosa
   KohonenMap.AddTrainPattern(5.3,3.7,1.5,0.2); // setosa
   KohonenMap.AddTrainPattern(5.0,3.3,1.4,0.2); // setosa
   KohonenMap.AddTrainPattern(7.0,3.2,4.7,1.4); // versicolor
   KohonenMap.AddTrainPattern(6.4,3.2,4.5,1.5); // versicolor
   KohonenMap.AddTrainPattern(6.9,3.1,4.9,1.5); // versicolor
   KohonenMap.AddTrainPattern(5.5,2.3,4.0,1.3); // versicolor
   KohonenMap.AddTrainPattern(6.5,2.8,4.6,1.5); // versicolor
   KohonenMap.AddTrainPattern(5.7,2.8,4.5,1.3); // versicolor
   KohonenMap.AddTrainPattern(6.3,3.3,4.7,1.6); // versicolor
   KohonenMap.AddTrainPattern(4.9,2.4,3.3,1.0); // versicolor
   KohonenMap.AddTrainPattern(6.6,2.9,4.6,1.3); // versicolor
   KohonenMap.AddTrainPattern(5.2,2.7,3.9,1.4); // versicolor
   KohonenMap.AddTrainPattern(5.0,2.0,3.5,1.0); // versicolor
   KohonenMap.AddTrainPattern(5.9,3.0,4.2,1.5); // versicolor
   KohonenMap.AddTrainPattern(6.0,2.2,4.0,1.0); // versicolor
   KohonenMap.AddTrainPattern(6.1,2.9,4.7,1.4); // versicolor
   KohonenMap.AddTrainPattern(5.6,2.9,3.6,1.3); // versicolor
   KohonenMap.AddTrainPattern(6.7,3.1,4.4,1.4); // versicolor
   KohonenMap.AddTrainPattern(5.6,3.0,4.5,1.5); // versicolor
   KohonenMap.AddTrainPattern(5.8,2.7,4.1,1.0); // versicolor
   KohonenMap.AddTrainPattern(6.2,2.2,4.5,1.5); // versicolor
   KohonenMap.AddTrainPattern(5.6,2.5,3.9,1.1); // versicolor
   KohonenMap.AddTrainPattern(5.9,3.2,4.8,1.8); // versicolor
   KohonenMap.AddTrainPattern(6.1,2.8,4.0,1.3); // versicolor
   KohonenMap.AddTrainPattern(6.3,2.5,4.9,1.5); // versicolor
   KohonenMap.AddTrainPattern(6.1,2.8,4.7,1.2); // versicolor
   KohonenMap.AddTrainPattern(6.4,2.9,4.3,1.3); // versicolor
   KohonenMap.AddTrainPattern(6.6,3.0,4.4,1.4); // versicolor
   KohonenMap.AddTrainPattern(6.8,2.8,4.8,1.4); // versicolor
   KohonenMap.AddTrainPattern(6.7,3.0,5.0,1.7); // versicolor
   KohonenMap.AddTrainPattern(6.0,2.9,4.5,1.5); // versicolor
   KohonenMap.AddTrainPattern(5.7,2.6,3.5,1.0); // versicolor
   KohonenMap.AddTrainPattern(5.5,2.4,3.8,1.1); // versicolor
   KohonenMap.AddTrainPattern(5.5,2.4,3.7,1.0); // versicolor
   KohonenMap.AddTrainPattern(5.8,2.7,3.9,1.2); // versicolor
   KohonenMap.AddTrainPattern(6.0,2.7,5.1,1.6); // versicolor
   KohonenMap.AddTrainPattern(5.4,3.0,4.5,1.5); // versicolor
   KohonenMap.AddTrainPattern(6.0,3.4,4.5,1.6); // versicolor
   KohonenMap.AddTrainPattern(6.7,3.1,4.7,1.5); // versicolor
   KohonenMap.AddTrainPattern(6.3,2.3,4.4,1.3); // versicolor
   KohonenMap.AddTrainPattern(5.6,3.0,4.1,1.3); // versicolor
   KohonenMap.AddTrainPattern(5.5,2.5,4.0,1.3); // versicolor
   KohonenMap.AddTrainPattern(5.5,2.6,4.4,1.2); // versicolor
   KohonenMap.AddTrainPattern(6.1,3.0,4.6,1.4); // versicolor
   KohonenMap.AddTrainPattern(5.8,2.6,4.0,1.2); // versicolor
   KohonenMap.AddTrainPattern(5.0,2.3,3.3,1.0); // versicolor
   KohonenMap.AddTrainPattern(5.6,2.7,4.2,1.3); // versicolor
   KohonenMap.AddTrainPattern(5.7,3.0,4.2,1.2); // versicolor
   KohonenMap.AddTrainPattern(5.7,2.9,4.2,1.3); // versicolor
   KohonenMap.AddTrainPattern(6.2,2.9,4.3,1.3); // versicolor
   KohonenMap.AddTrainPattern(5.1,2.5,3.0,1.1); // versicolor
   KohonenMap.AddTrainPattern(5.7,2.8,4.1,1.3); // versicolor
   KohonenMap.AddTrainPattern(6.3,3.3,6.0,2.5); // virginica
   KohonenMap.AddTrainPattern(5.8,2.7,5.1,1.9); // virginica
   KohonenMap.AddTrainPattern(7.1,3.0,5.9,2.1); // virginica
   KohonenMap.AddTrainPattern(6.3,2.9,5.6,1.8); // virginica
   KohonenMap.AddTrainPattern(6.5,3.0,5.8,2.2); // virginica
   KohonenMap.AddTrainPattern(7.6,3.0,6.6,2.1); // virginica
   KohonenMap.AddTrainPattern(4.9,2.5,4.5,1.7); // virginica
   KohonenMap.AddTrainPattern(7.3,2.9,6.3,1.8); // virginica
   KohonenMap.AddTrainPattern(6.7,2.5,5.8,1.8); // virginica
   KohonenMap.AddTrainPattern(7.2,3.6,6.1,2.5); // virginica
   KohonenMap.AddTrainPattern(6.5,3.2,5.1,2.0); // virginica
   KohonenMap.AddTrainPattern(6.4,2.7,5.3,1.9); // virginica
   KohonenMap.AddTrainPattern(6.8,3.0,5.5,2.1); // virginica
   KohonenMap.AddTrainPattern(5.7,2.5,5.0,2.0); // virginica
   KohonenMap.AddTrainPattern(5.8,2.8,5.1,2.4); // virginica
   KohonenMap.AddTrainPattern(6.4,3.2,5.3,2.3); // virginica
   KohonenMap.AddTrainPattern(6.5,3.0,5.5,1.8); // virginica
   KohonenMap.AddTrainPattern(7.7,3.8,6.7,2.2); // virginica
   KohonenMap.AddTrainPattern(7.7,2.6,6.9,2.3); // virginica
   KohonenMap.AddTrainPattern(6.0,2.2,5.0,1.5); // virginica
   KohonenMap.AddTrainPattern(6.9,3.2,5.7,2.3); // virginica
   KohonenMap.AddTrainPattern(5.6,2.8,4.9,2.0); // virginica
   KohonenMap.AddTrainPattern(7.7,2.8,6.7,2.0); // virginica
   KohonenMap.AddTrainPattern(6.3,2.7,4.9,1.8); // virginica
   KohonenMap.AddTrainPattern(6.7,3.3,5.7,2.1); // virginica
   KohonenMap.AddTrainPattern(7.2,3.2,6.0,1.8); // virginica
   KohonenMap.AddTrainPattern(6.2,2.8,4.8,1.8); // virginica
   KohonenMap.AddTrainPattern(6.1,3.0,4.9,1.8); // virginica
   KohonenMap.AddTrainPattern(6.4,2.8,5.6,2.1); // virginica
   KohonenMap.AddTrainPattern(7.2,3.0,5.8,1.6); // virginica
   KohonenMap.AddTrainPattern(7.4,2.8,6.1,1.9); // virginica
   KohonenMap.AddTrainPattern(7.9,3.8,6.4,2.0); // virginica
   KohonenMap.AddTrainPattern(6.4,2.8,5.6,2.2); // virginica
   KohonenMap.AddTrainPattern(6.3,2.8,5.1,1.5); // virginica
   KohonenMap.AddTrainPattern(6.1,2.6,5.6,1.4); // virginica
   KohonenMap.AddTrainPattern(7.7,3.0,6.1,2.3); // virginica
   KohonenMap.AddTrainPattern(6.3,3.4,5.6,2.4); // virginica
   KohonenMap.AddTrainPattern(6.4,3.1,5.5,1.8); // virginica
   KohonenMap.AddTrainPattern(6.0,3.0,4.8,1.8); // virginica
   KohonenMap.AddTrainPattern(6.9,3.1,5.4,2.1); // virginica
   KohonenMap.AddTrainPattern(6.7,3.1,5.6,2.4); // virginica
   KohonenMap.AddTrainPattern(6.9,3.1,5.1,2.3); // virginica
   KohonenMap.AddTrainPattern(5.8,2.7,5.1,1.9); // virginica
   KohonenMap.AddTrainPattern(6.8,3.2,5.9,2.3); // virginica
   KohonenMap.AddTrainPattern(6.7,3.3,5.7,2.5); // virginica
   KohonenMap.AddTrainPattern(6.7,3.0,5.2,2.3); // virginica
   KohonenMap.AddTrainPattern(6.3,2.5,5.0,1.9); // virginica
   KohonenMap.AddTrainPattern(6.5,3.0,5.2,2.0); // virginica
   KohonenMap.AddTrainPattern(6.2,3.4,5.4,2.3); // virginica
   KohonenMap.AddTrainPattern(5.9,3.0,5.1,1.8); // virginica

//---  
   KohonenMap.Train();
//---    
   KohonenMap.Render();
//---          
   KohonenMap.ShowPattern(5.1,3.5,1.4,0.2,"setosa");
   KohonenMap.ShowPattern(4.9,3.0,1.4,0.2,"setosa");
   KohonenMap.ShowPattern(4.7,3.2,1.3,0.2,"setosa");
   KohonenMap.ShowPattern(4.6,3.1,1.5,0.2,"setosa");
   KohonenMap.ShowPattern(5.0,3.6,1.4,0.2,"setosa");
   KohonenMap.ShowPattern(5.4,3.9,1.7,0.4,"setosa");
   KohonenMap.ShowPattern(4.6,3.4,1.4,0.3,"setosa");
   KohonenMap.ShowPattern(5.0,3.4,1.5,0.2,"setosa");
   KohonenMap.ShowPattern(4.4,2.9,1.4,0.2,"setosa");
   KohonenMap.ShowPattern(4.9,3.1,1.5,0.1,"setosa");
   KohonenMap.ShowPattern(5.4,3.7,1.5,0.2,"setosa");
   KohonenMap.ShowPattern(4.8,3.4,1.6,0.2,"setosa");
   KohonenMap.ShowPattern(4.8,3.0,1.4,0.1,"setosa");
   KohonenMap.ShowPattern(4.3,3.0,1.1,0.1,"setosa");
   KohonenMap.ShowPattern(5.8,4.0,1.2,0.2,"setosa");
   KohonenMap.ShowPattern(5.7,4.4,1.5,0.4,"setosa");
   KohonenMap.ShowPattern(5.4,3.9,1.3,0.4,"setosa");
   KohonenMap.ShowPattern(5.1,3.5,1.4,0.3,"setosa");
   KohonenMap.ShowPattern(5.7,3.8,1.7,0.3,"setosa");
   KohonenMap.ShowPattern(5.1,3.8,1.5,0.3,"setosa");
   KohonenMap.ShowPattern(5.4,3.4,1.7,0.2,"setosa");
   KohonenMap.ShowPattern(5.1,3.7,1.5,0.4,"setosa");
   KohonenMap.ShowPattern(4.6,3.6,1.0,0.2,"setosa");
   KohonenMap.ShowPattern(5.1,3.3,1.7,0.5,"setosa");
   KohonenMap.ShowPattern(4.8,3.4,1.9,0.2,"setosa");
   KohonenMap.ShowPattern(5.0,3.0,1.6,0.2,"setosa");
   KohonenMap.ShowPattern(5.0,3.4,1.6,0.4,"setosa");
   KohonenMap.ShowPattern(5.2,3.5,1.5,0.2,"setosa");
   KohonenMap.ShowPattern(5.2,3.4,1.4,0.2,"setosa");
   KohonenMap.ShowPattern(4.7,3.2,1.6,0.2,"setosa");
   KohonenMap.ShowPattern(4.8,3.1,1.6,0.2,"setosa");
   KohonenMap.ShowPattern(5.4,3.4,1.5,0.4,"setosa");
   KohonenMap.ShowPattern(5.2,4.1,1.5,0.1,"setosa");
   KohonenMap.ShowPattern(5.5,4.2,1.4,0.2,"setosa");
   KohonenMap.ShowPattern(4.9,3.1,1.5,0.2,"setosa");
   KohonenMap.ShowPattern(5.0,3.2,1.2,0.2,"setosa");
   KohonenMap.ShowPattern(5.5,3.5,1.3,0.2,"setosa");
   KohonenMap.ShowPattern(4.9,3.6,1.4,0.1,"setosa");
   KohonenMap.ShowPattern(4.4,3.0,1.3,0.2,"setosa");
   KohonenMap.ShowPattern(5.1,3.4,1.5,0.2,"setosa");
   KohonenMap.ShowPattern(5.0,3.5,1.3,0.3,"setosa");
   KohonenMap.ShowPattern(4.5,2.3,1.3,0.3,"setosa");
   KohonenMap.ShowPattern(4.4,3.2,1.3,0.2,"setosa");
   KohonenMap.ShowPattern(5.0,3.5,1.6,0.6,"setosa");
   KohonenMap.ShowPattern(5.1,3.8,1.9,0.4,"setosa");
   KohonenMap.ShowPattern(4.8,3.0,1.4,0.3,"setosa");
   KohonenMap.ShowPattern(5.1,3.8,1.6,0.2,"setosa");
   KohonenMap.ShowPattern(4.6,3.2,1.4,0.2,"setosa");
   KohonenMap.ShowPattern(5.3,3.7,1.5,0.2,"setosa");
   KohonenMap.ShowPattern(5.0,3.3,1.4,0.2,"setosa");
   KohonenMap.ShowPattern(7.0,3.2,4.7,1.4,"versicolor");
   KohonenMap.ShowPattern(6.4,3.2,4.5,1.5,"versicolor");
   KohonenMap.ShowPattern(6.9,3.1,4.9,1.5,"versicolor");
   KohonenMap.ShowPattern(5.5,2.3,4.0,1.3,"versicolor");
   KohonenMap.ShowPattern(6.5,2.8,4.6,1.5,"versicolor");
   KohonenMap.ShowPattern(5.7,2.8,4.5,1.3,"versicolor");
   KohonenMap.ShowPattern(6.3,3.3,4.7,1.6,"versicolor");
   KohonenMap.ShowPattern(4.9,2.4,3.3,1.0,"versicolor");
   KohonenMap.ShowPattern(6.6,2.9,4.6,1.3,"versicolor");
   KohonenMap.ShowPattern(5.2,2.7,3.9,1.4,"versicolor");
   KohonenMap.ShowPattern(5.0,2.0,3.5,1.0,"versicolor");
   KohonenMap.ShowPattern(5.9,3.0,4.2,1.5,"versicolor");
   KohonenMap.ShowPattern(6.0,2.2,4.0,1.0,"versicolor");
   KohonenMap.ShowPattern(6.1,2.9,4.7,1.4,"versicolor");
   KohonenMap.ShowPattern(5.6,2.9,3.6,1.3,"versicolor");
   KohonenMap.ShowPattern(6.7,3.1,4.4,1.4,"versicolor");
   KohonenMap.ShowPattern(5.6,3.0,4.5,1.5,"versicolor");
   KohonenMap.ShowPattern(5.8,2.7,4.1,1.0,"versicolor");
   KohonenMap.ShowPattern(6.2,2.2,4.5,1.5,"versicolor");
   KohonenMap.ShowPattern(5.6,2.5,3.9,1.1,"versicolor");
   KohonenMap.ShowPattern(5.9,3.2,4.8,1.8,"versicolor");
   KohonenMap.ShowPattern(6.1,2.8,4.0,1.3,"versicolor");
   KohonenMap.ShowPattern(6.3,2.5,4.9,1.5,"versicolor");
   KohonenMap.ShowPattern(6.1,2.8,4.7,1.2,"versicolor");
   KohonenMap.ShowPattern(6.4,2.9,4.3,1.3,"versicolor");
   KohonenMap.ShowPattern(6.6,3.0,4.4,1.4,"versicolor");
   KohonenMap.ShowPattern(6.8,2.8,4.8,1.4,"versicolor");
   KohonenMap.ShowPattern(6.7,3.0,5.0,1.7,"versicolor");
   KohonenMap.ShowPattern(6.0,2.9,4.5,1.5,"versicolor");
   KohonenMap.ShowPattern(5.7,2.6,3.5,1.0,"versicolor");
   KohonenMap.ShowPattern(5.5,2.4,3.8,1.1,"versicolor");
   KohonenMap.ShowPattern(5.5,2.4,3.7,1.0,"versicolor");
   KohonenMap.ShowPattern(5.8,2.7,3.9,1.2,"versicolor");
   KohonenMap.ShowPattern(6.0,2.7,5.1,1.6,"versicolor");
   KohonenMap.ShowPattern(5.4,3.0,4.5,1.5,"versicolor");
   KohonenMap.ShowPattern(6.0,3.4,4.5,1.6,"versicolor");
   KohonenMap.ShowPattern(6.7,3.1,4.7,1.5,"versicolor");
   KohonenMap.ShowPattern(6.3,2.3,4.4,1.3,"versicolor");
   KohonenMap.ShowPattern(5.6,3.0,4.1,1.3,"versicolor");
   KohonenMap.ShowPattern(5.5,2.5,4.0,1.3,"versicolor");
   KohonenMap.ShowPattern(5.5,2.6,4.4,1.2,"versicolor");
   KohonenMap.ShowPattern(6.1,3.0,4.6,1.4,"versicolor");
   KohonenMap.ShowPattern(5.8,2.6,4.0,1.2,"versicolor");
   KohonenMap.ShowPattern(5.0,2.3,3.3,1.0,"versicolor");
   KohonenMap.ShowPattern(5.6,2.7,4.2,1.3,"versicolor");
   KohonenMap.ShowPattern(5.7,3.0,4.2,1.2,"versicolor");
   KohonenMap.ShowPattern(5.7,2.9,4.2,1.3,"versicolor");
   KohonenMap.ShowPattern(6.2,2.9,4.3,1.3,"versicolor");
   KohonenMap.ShowPattern(5.1,2.5,3.0,1.1,"versicolor");
   KohonenMap.ShowPattern(5.7,2.8,4.1,1.3,"versicolor");
   KohonenMap.ShowPattern(6.3,3.3,6.0,2.5,"virginica");
   KohonenMap.ShowPattern(5.8,2.7,5.1,1.9,"virginica");
   KohonenMap.ShowPattern(7.1,3.0,5.9,2.1,"virginica");
   KohonenMap.ShowPattern(6.3,2.9,5.6,1.8,"virginica");
   KohonenMap.ShowPattern(6.5,3.0,5.8,2.2,"virginica");
   KohonenMap.ShowPattern(7.6,3.0,6.6,2.1,"virginica");
   KohonenMap.ShowPattern(4.9,2.5,4.5,1.7,"virginica");
   KohonenMap.ShowPattern(7.3,2.9,6.3,1.8,"virginica");
   KohonenMap.ShowPattern(6.7,2.5,5.8,1.8,"virginica");
   KohonenMap.ShowPattern(7.2,3.6,6.1,2.5,"virginica");
   KohonenMap.ShowPattern(6.5,3.2,5.1,2.0,"virginica");
   KohonenMap.ShowPattern(6.4,2.7,5.3,1.9,"virginica");
   KohonenMap.ShowPattern(6.8,3.0,5.5,2.1,"virginica");
   KohonenMap.ShowPattern(5.7,2.5,5.0,2.0,"virginica");
   KohonenMap.ShowPattern(5.8,2.8,5.1,2.4,"virginica");
   KohonenMap.ShowPattern(6.4,3.2,5.3,2.3,"virginica");
   KohonenMap.ShowPattern(6.5,3.0,5.5,1.8,"virginica");
   KohonenMap.ShowPattern(7.7,3.8,6.7,2.2,"virginica");
   KohonenMap.ShowPattern(7.7,2.6,6.9,2.3,"virginica");
   KohonenMap.ShowPattern(6.0,2.2,5.0,1.5,"virginica");
   KohonenMap.ShowPattern(6.9,3.2,5.7,2.3,"virginica");
   KohonenMap.ShowPattern(5.6,2.8,4.9,2.0,"virginica");
   KohonenMap.ShowPattern(7.7,2.8,6.7,2.0,"virginica");
   KohonenMap.ShowPattern(6.3,2.7,4.9,1.8,"virginica");
   KohonenMap.ShowPattern(6.7,3.3,5.7,2.1,"virginica");
   KohonenMap.ShowPattern(7.2,3.2,6.0,1.8,"virginica");
   KohonenMap.ShowPattern(6.2,2.8,4.8,1.8,"virginica");
   KohonenMap.ShowPattern(6.1,3.0,4.9,1.8,"virginica");
   KohonenMap.ShowPattern(6.4,2.8,5.6,2.1,"virginica");
   KohonenMap.ShowPattern(7.2,3.0,5.8,1.6,"virginica");
   KohonenMap.ShowPattern(7.4,2.8,6.1,1.9,"virginica");
   KohonenMap.ShowPattern(7.9,3.8,6.4,2.0,"virginica");
   KohonenMap.ShowPattern(6.4,2.8,5.6,2.2,"virginica");
   KohonenMap.ShowPattern(6.3,2.8,5.1,1.5,"virginica");
   KohonenMap.ShowPattern(6.1,2.6,5.6,1.4,"virginica");
   KohonenMap.ShowPattern(7.7,3.0,6.1,2.3,"virginica");
   KohonenMap.ShowPattern(6.3,3.4,5.6,2.4,"virginica");
   KohonenMap.ShowPattern(6.4,3.1,5.5,1.8,"virginica");
   KohonenMap.ShowPattern(6.0,3.0,4.8,1.8,"virginica");
   KohonenMap.ShowPattern(6.9,3.1,5.4,2.1,"virginica");
   KohonenMap.ShowPattern(6.7,3.1,5.6,2.4,"virginica");
   KohonenMap.ShowPattern(6.9,3.1,5.1,2.3,"virginica");
   KohonenMap.ShowPattern(5.8,2.7,5.1,1.9,"virginica");
   KohonenMap.ShowPattern(6.8,3.2,5.9,2.3,"virginica");
   KohonenMap.ShowPattern(6.7,3.3,5.7,2.5,"virginica");
   KohonenMap.ShowPattern(6.7,3.0,5.2,2.3,"virginica");
   KohonenMap.ShowPattern(6.3,2.5,5.0,1.9,"virginica");
   KohonenMap.ShowPattern(6.5,3.0,5.2,2.0,"virginica");
   KohonenMap.ShowPattern(6.2,3.4,5.4,2.3,"virginica");
   KohonenMap.ShowPattern(5.9,3.0,5.1,1.8,"virginica");

//---    
   KohonenMap.ShowBMP(false);
  }
//+------------------------------------------------------------------+
//| Expert deinitialization function                                 |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
   KohonenMap.NetDeinit();
  }
//+------------------------------------------------------------------+
//| Expert tick function                                             |
//+------------------------------------------------------------------+
void OnTick()
  {
//---
  }
//+------------------------------------------------------------------+
