//+------------------------------------------------------------------+
//|                                                      SOM_ex5.mq5 |
//|                        Copyright 2011, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2011, MetaQuotes Software Corp."
#property link      "http://www.mql5.com"
#property version   "1.00"

enum ColorSchemes
  {
   Black_White=0,      // -
   Blue_Green_Red=1,   // ----
   Red_Black_Green=2,  // --
   Red_White_Green=3   // --
  };

//  ,      
input bool HexagonalCell=true;
//  ,    
input bool ShowBorders=false;
//  ,      
input bool ShowTitles=true;
//  ,    
input ColorSchemes ColorScheme=Red_White_Green;
//  ,  . -   
input int MaxPictures=4;

//  ,     X
input int CellsX=30;
//  ,     Y
input int CellsY=30;

//  ,   
input int ImageXSize=250;
//  ,   
input int ImageYSize=250;
//  ,    
input string DataFileName="optim.csv";

//---   cIntBMP
#include <cIntBMP.mqh>
//+------------------------------------------------------------------+
//|   SOM-                                              |
//+------------------------------------------------------------------+
class CSOMNode
  {
protected:
   int               m_x1;
   int               m_y1;
   int               m_x2;
   int               m_y2;
   int               m_dimension;

   double            m_x;
   double            m_y;
   double            m_weights[];
public:
   //---  
                     CSOMNode();
   //---  
                    ~CSOMNode();
   //---    
   void              InitNode(int x1,int y1,int x2,int y2);
   //---     
   double            X()  const { return(m_x);}
   double            Y()  const { return(m_y);}
   //---    
   void              GetCoordinates(int &x1,int &y1,int &x2,int &y2);
   //---     weight_index
   double            GetWeight(int weight_index);
   //---       
   double            CalculateDistance(double &vector[]);
   //---    
   void              AdjustWeights(double &vector[],double learning_rate,double influence);
   //---    
   void              SetDimension(int dimension);

  };
//+------------------------------------------------------------------+
//|                                                 |
//+------------------------------------------------------------------+
void CSOMNode::CSOMNode()
  {
   m_x1=0;
   m_y1=0;
   m_x2=0;
   m_y2=0;
   m_x=0.0;
   m_y=0.0;
  };
//+------------------------------------------------------------------+
//|                                       |
//+------------------------------------------------------------------+
void CSOMNode::InitNode(int x1,int y1,int x2,int y2)
  {
//---   
   m_x1=x1;
   m_y1=y1;
   m_x2=x2;
   m_y2=y2;
//---    
   m_x=x1+MathAbs((x2-x1))/2;
   m_y=y1+MathAbs((y2-y1))/2;
//---      
   ArrayResize(m_weights,m_dimension);
//---    ;
   for(int i=0; i<m_dimension; i++) {m_weights[i]=255.0*rand()/32768;}
  };
//+------------------------------------------------------------------+
//|                                                  |
//+------------------------------------------------------------------+
void CSOMNode::~CSOMNode()
  {
   ArrayResize(m_weights,0);
  };
//+------------------------------------------------------------------+
//|     weight_index                    |
//+------------------------------------------------------------------+
double CSOMNode::GetWeight(int weight_index)
  {
   if(weight_index>=0 && weight_index<m_dimension)
     {
      return(m_weights[weight_index]);
     }
   else return(0);
  }
//+------------------------------------------------------------------+
//|                                   |
//+------------------------------------------------------------------+
void CSOMNode::GetCoordinates(int &x1,int &y1,int &x2,int &y2)
  {
   x1=m_x1;
   y1=m_y1;
   x2=m_x2;
   y2=m_y2;
  }
//+------------------------------------------------------------------+
//|              |
//+------------------------------------------------------------------+
double CSOMNode::CalculateDistance(double &vector[])
  {
   double dist_sqr=0;
   for(int i=0; i<m_dimension; i++)
     {
      dist_sqr=dist_sqr+(vector[i]-m_weights[i])*(vector[i]-m_weights[i]);
     }
   return(dist_sqr);
  };
//+------------------------------------------------------------------+
//|                |
//+------------------------------------------------------------------+
void CSOMNode::AdjustWeights(double &vector[],double learning_rate,double influence)
  {
   for(int i=0; i<m_dimension; i++)
     {
      m_weights[i]=m_weights[i]+learning_rate*influence*(vector[i]-m_weights[i]);
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOMNode::SetDimension(int dimension)
  {
   m_dimension=dimension;
  }
//+------------------------------------------------------------------+
//|  SOM                                                        |
//+------------------------------------------------------------------+
class CSOM
  {
protected:
   uchar             Palette[768];
   //---      BMP-
   cIntBMP           m_bmp[];
   //---      
   int               m_gridmode;
   //---  BMP-
   int               m_xsize;
   int               m_ysize;
   //---  
   int               m_xcells;
   int               m_ycells;
   //---    
   CSOMNode          m_som_nodes[];
   //---    
   int               m_total_training_sets;
   //---    
   double            m_training_sets_array[];
   double            m_max_values[];
   double            m_min_values[];

protected:
   int               m_dimension;
   double            m_map_radius;
   double            m_time_constant;

   double            m_initial_learning_rate;
   int               m_iterations;
   string            m_som_titles[];
   string            m_train_titles[];

public:
   //---  
                     CSOM();
   //---  
                    ~CSOM();
   //---   
   void              InitParameters(int dimension,int iterations,int xcells,int ycells,int bmpwidth,int bmpheight);
   //---        
   int               BestMatchingNode(double &vector[]);
   //---   
   void              Train();
   //---    
   void              Render();
   //---   bmp-  
   void              ShowBMP(bool back);
   //---      
   void              AddVectorToTrainingSet(double &vector[],string title);
   //---       
   void              ShowPattern(double &vector[],string name);
   //---    col  RGB-
   void              ColToRGB(int col,int &r,int &g,int &b);
   //---     RGB-
   int               RGB256(int r,int g,int b) const {return(r+256*g+65536*b);      }
   //---       
   int               GetPalColor(uchar index);
   //---    
   void              Blend(uchar c1,uchar c2,uchar r1,uchar g1,uchar b1,uchar r2,uchar g2,uchar b2);
   //---   
   void              RenderCell(int img,int col,int node_index,bool cell_even);
   //---   ,  BMP-  
   void              NetDeinit();
   //---    CSV-
   void              ParseString(string value,string delimiter,string  &mystrings[]);
   //---        
   bool              LoadTrainDataFromFile(string filename);
   //---     CSV-
   void              ReadCSVData(int handle);
   //---       
   void              ShowTrainPatterns();
  };
//+------------------------------------------------------------------+
//|                                                       |
//+------------------------------------------------------------------+
void CSOM::CSOM()
  {
   m_iterations=10;
   m_initial_learning_rate=0.1;
   m_total_training_sets=0;
   ArrayResize(m_training_sets_array,0);
   ArrayResize(m_som_titles,0);
   ArrayResize(m_train_titles,0);
  }
//+------------------------------------------------------------------+
//|                                                        |
//+------------------------------------------------------------------+
void CSOM::~CSOM()
  {
   ArrayResize(m_training_sets_array,0);
   ArrayResize(m_som_titles,0);
   ArrayResize(m_train_titles,0);
  }
//+------------------------------------------------------------------+
//|                                         |
//+------------------------------------------------------------------+
void CSOM::InitParameters(int dimension,int iterations,int xcells,int ycells,int bmpwidth,int bmpheight)
  {
   m_dimension=dimension;
   ArrayResize(m_som_titles,m_dimension);
   for(int i=0; i<m_dimension; i++)
     {
      m_som_titles[i]="Param"+IntegerToString(i+1);
     }
   m_xsize=bmpwidth;
   m_ysize=bmpheight;

   m_xcells=xcells;
   m_ycells=ycells;

//---       
   m_gridmode=int(ChartGetInteger(0,CHART_SHOW_GRID,0));

//---      
   m_iterations=iterations;
   int cellwidth=m_xsize/m_xcells;
   int cellheight=m_ysize/m_ycells;

//---  BMP-
   ArrayResize(m_bmp,m_dimension+1);
   for(int i=0; i<m_dimension; i++)
     {
      if(!HexagonalCell) {m_bmp[i].Create(m_xsize,m_ysize,clrWhite);}
      else m_bmp[i].Create(m_xsize+cellwidth,m_ysize+cellheight/2,clrWhite);
     }

//default
   Blend(0,255,0,0,0,255,255,255);

   switch(ColorScheme)
     {
      case 0:
        {
         //--- Black to White gradient
         Blend(0,255,0,0,0,255,255,255);
         break;
        }

      case 1:
        {//--- Blue-green-red
         Blend(0,1*64-1,98,98,255,98,255,255);
         Blend(1*64,2*64-1,98,255,255,98,255,98);
         Blend(2*64,3*64-1,98,255,98,255,255,98);
         Blend(3*64,4*64-1,255,255,98,255,98,98);
         break;
        }

      case 2:
        {//--- Red-black-green
         Blend(0,127,155,0,0,0,0,0);
         Blend(127,255,0,0,0,0,155,0);
         break;
        }
      case 3:
        {//--- Red-white-green
         Blend(0,127,155,0,0,255,255,255);
         Blend(127,255,255,255,255,0,155,0);
         break;
        }
     }

/* Blend(0,1*64-1,98,98,255,98,255,255);
   Blend(1*64,2*64-1,98,255,255,98,255,98);
   Blend(2*64,3*64-1,98,255,98,255,255,98);
   Blend(3*64,4*64-1,255,255,98,255,98,98);
   */

//---        
   ArrayResize(m_som_nodes,xcells*ycells);
   int ind=0;
//---  
   for(int i=0; i<m_xcells; i++)
     {
      for(int j=0; j<m_ycells; j++)
        {
         m_som_nodes[ind].SetDimension(m_dimension);
         m_som_nodes[ind].InitNode(i*cellwidth,j*cellheight,(i+1)*cellwidth,(j+1)*cellheight);
         ind++;
        }
     }
//---    
   if(m_xsize>m_ysize) {m_map_radius=m_xsize/2.0;}
   else {m_map_radius=m_ysize/2.0;}
//Print("m_map_radius=",m_map_radius);
   m_time_constant=1.0*m_iterations/MathLog(m_map_radius);
   Print("m_time_constant=",m_time_constant);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM::Blend(uchar c1,uchar c2,uchar r1,uchar g1,uchar b1,uchar r2,uchar g2,uchar b2)
  {
   int n=(c2-c1);
   for(int i=0; i<=n; i++)
     {
      if((c1+i+2)<ArraySize(Palette))
        {
         Palette[3*(c1+i)]=uchar(MathRound(1*(r1*(n-i)+r2*i)*1.0/n));
         Palette[3*(c1+i)+1]=uchar (MathRound(1*(g1*(n-i) + g2*i) * 1.0/ n));
         Palette[3*(c1+i)+2]=uchar (MathRound(1*(b1*(n-i) + b2*i) * 1.0/ n));
        }
     }
  }
//+------------------------------------------------------------------+
//|                          |
//+------------------------------------------------------------------+
int CSOM::GetPalColor(uchar index)
  {
   int ind=index;
   if(ind<=0) {ind=0;}
   uchar r=Palette[3*(ind)];
   uchar g=Palette[3*(ind)+1];
   uchar b=Palette[3*(ind)+2];
   return(RGB256(r,g,b));
  }
//+------------------------------------------------------------------+
//|                              |
//+------------------------------------------------------------------+
int CSOM::BestMatchingNode(double &vector[])
  {
   int    min_ind=0;
   double min_dist=m_som_nodes[min_ind].CalculateDistance(vector);
   int total_nodes= ArraySize(m_som_nodes);
   for(int i=0; i<total_nodes; i++)
     {
      if(m_som_nodes[i].CalculateDistance(vector)<min_dist)
        {
         min_dist=m_som_nodes[i].CalculateDistance(vector);
         min_ind=i;
        };
     }
   return(min_ind);
  }
//+------------------------------------------------------------------+
//|                                              |
//+------------------------------------------------------------------+
void CSOM::Train()
  {
   double learning_rate=m_initial_learning_rate;
   int iter=0;
   double datavector[];
   ArrayResize(datavector,m_dimension);

   Print("Dimension=",m_dimension);

//---        
   ArrayResize(m_max_values,m_dimension);
   ArrayResize(m_min_values,m_dimension);

   for(int j=0; j<m_dimension; j++)
     {
      double maxv=m_training_sets_array[m_dimension+j];
      double minv=m_training_sets_array[m_dimension+j];
      for(int i=1; i<m_total_training_sets; i++)
        {
         double v=m_training_sets_array[m_dimension*i+j];
         if(v>maxv) {maxv=v;}
         if(v<minv) {minv=v;}
        }
      m_max_values[j]=maxv;
      m_min_values[j]=minv;
      Print(j,"m_min_value=",m_min_values[j],"m_max_value=",m_max_values[j]);
     }

   int total_nodes=ArraySize(m_som_nodes);

   int frame=0;
   do
     {
      //---        
      int ind=int(1.0*m_total_training_sets*rand()/32768);

      //---  datavector  

      for(int k=0; k<m_dimension; k++)
        {
         datavector[k]=m_training_sets_array[m_dimension*(ind)+k];
        }

      //---    ,    datavector
      int winningnode=BestMatchingNode(datavector);
      //---    
      double neighbourhood_radius=m_map_radius*MathExp(-1.0*iter/m_time_constant);

      //---     
      for(int i=0; i<total_nodes; i++)
        {
         //---        m_WinningNodeIndex
         double DistToNodeSqr=(m_som_nodes[winningnode].X()-m_som_nodes[i].X())*(m_som_nodes[winningnode].X()-m_som_nodes[i].X())
                              +(m_som_nodes[winningnode].Y()-m_som_nodes[i].Y())*(m_som_nodes[winningnode].Y()-m_som_nodes[i].Y());
         //---                                 
         double WS=neighbourhood_radius*neighbourhood_radius;
         //---    ,   
         if(DistToNodeSqr<WS)
           {
            //---     
            double influence=MathExp(-DistToNodeSqr/(2*WS));
            //---       
            m_som_nodes[i].AdjustWeights(datavector,learning_rate,influence);
           }
        }
      //---  ( iter)   
      learning_rate=m_initial_learning_rate*MathExp(-1.0*iter/m_iterations);
      //---   
      iter++;
      //---  ,    
      //---      N- 
      if(iter%100==0)
        {
         frame++;
         Print(" ",iter,"  ",m_iterations," ",int(iter*100.0/m_iterations),"%");
         Render();
         ShowBMP(false);
        }
     }
   //---     ,       
   while(iter<m_iterations);

   Print(" .");
  }//+------------------------------------------------------------------+
//|   bmp-                 |
//+------------------------------------------------------------------+
void CSOM::ShowBMP(bool back)
  {
   for(int i=0; i<m_dimension; i++)
     {
      //---  BMP-  
      m_bmp[i].Save("bmpimg"+IntegerToString(i),true);
      //---    
      int pos1=i/MaxPictures;
      int pos2=i%MaxPictures;
      //m_bmp[i].Show(10+i*(15+m_xsize),10+0,"bmpimg"+IntegerToString(i),"SOM"+IntegerToString(i));
      m_bmp[i].Show(10+pos2*(15+m_xsize),10+pos1*(15+m_ysize),"bmpimg"+IntegerToString(i),"SOM"+IntegerToString(i));
      //---    
      ObjectSetInteger(0,"SOM"+IntegerToString(i),OBJPROP_BACK,back);
     }
  }
//+------------------------------------------------------------------+
//|                                |
//+------------------------------------------------------------------+
void CSOM::AddVectorToTrainingSet(double &vector[],string title)
  {
   m_total_training_sets++;
   ArrayResize(m_training_sets_array,m_dimension*m_total_training_sets);
   ArrayResize(m_train_titles,m_total_training_sets);
   m_train_titles[m_total_training_sets-1]=title;
   for(int i=0; i<m_dimension; i++)
     {
      m_training_sets_array[m_dimension*(m_total_training_sets-1)+i]=vector[i];
     }
  }
//+------------------------------------------------------------------+
//|                       |
//+------------------------------------------------------------------+
void CSOM::ShowPattern(double &vector[],string name)
  {
   int WinningNodeIndex=BestMatchingNode(vector);
   int ind=BestMatchingNode(vector);

   int x1,y1,x2,y2;
   m_som_nodes[ind].GetCoordinates(x1,y1,x2,y2);

   int avgcolor=0;
   for(int i=0; i<m_dimension; i++)
     {
      avgcolor+=int((255*(m_som_nodes[ind].GetWeight(i)-m_min_values[i])/(m_max_values[i]-m_min_values[i])));
     }
   if(HexagonalCell)
     {
      //--- 
      int x_size=int (MathAbs(x2-x1));
      int y_size=int (MathAbs(y2-y1));
      y1=y1+y_size/4;
      y2=y2+y_size/4;
     }

   int xc=(x1+x2)/2;
   int yc=(y1+y2)/2;

   if(ShowTitles)
     {
      for(int i=0; i<m_dimension; i++)
        {
         if(avgcolor<=m_dimension*128)
           {
            m_bmp[i].TypeText(xc+1,yc+1,name,clrBlack);
            m_bmp[i].TypeText(xc,yc,name,clrWhite);
           }
         else
           {
            m_bmp[i].TypeText(xc+1,yc+1,name,clrBlack);
            m_bmp[i].TypeText(xc,yc,name,clrWhite);
           }
        }
     }

   Print(" ",name,":     =",ind);
  }
//+------------------------------------------------------------------+
//|         -      |
//+------------------------------------------------------------------+
void CSOM::ShowTrainPatterns()
  {
   double datavector[];
   ArrayResize(datavector,m_dimension);

   for(int ind=0; ind<m_total_training_sets; ind++)
     {
      //---  datavector     
      for(int k=0; k<m_dimension; k++)
        {
         datavector[k]=m_training_sets_array[m_dimension*(ind)+k];
        }
      ShowPattern(datavector,m_train_titles[ind]);
     }
  }
//+------------------------------------------------------------------+
//|    col  RGB-                    |
//+------------------------------------------------------------------+
void CSOM::ColToRGB(int col,int &r,int &g,int &b)
  {
   r=col & 0xFF;
   g=(col>>8)& 0xFF;
   b=(col>>16)& 0xFF;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM::RenderCell(int img,int col,int node_index,bool cell_even)
  {
   int x1,y1,x2,y2;
   m_som_nodes[node_index].GetCoordinates(x1,y1,x2,y2);

   if(!HexagonalCell)
     {
      //---  
      m_bmp[img].DrawRectangle(x1,y1,x2,y2,col,true);
      if(ShowBorders) m_bmp[img].DrawRectangle(x1,y1,x2,y2,clrBlack,false);
     }
   else
     {
      //--- 
      int x_size=int (MathAbs(x2-x1));
      int y_size=int (MathAbs(y2-y1));
      if(cell_even)
        {
         x1=x1+x_size/2;
         x2=x2+x_size/2;
        }
      y1=15+y1+y_size/4;
      y2=15+y2+y_size/4;

      m_bmp[img].DrawRectangle(x1,y1+y_size/4,x2,y2-y_size/4,col,true);
      m_bmp[img].DrawTriangle(x1,y1+y_size/4,x1+x_size/2,y1-y_size/4,x2,y1+y_size/4,col,true);
      m_bmp[img].DrawTriangle(x1,y2-y_size/4,x2,y2-y_size/4,x1+x_size/2,y2+y_size/4,col,true);

      if(ShowBorders)
        {
         //---  
         int bcol=clrBlack;
         m_bmp[img].DrawLine(x1,y1+y_size/4,x1+x_size/2,y1-y_size/4,bcol);
         m_bmp[img].DrawLine(x1+x_size/2,y1-y_size/4,x2,y1+y_size/4,bcol);

         m_bmp[img].DrawLine(x1,y2-y_size/4,x1+x_size/2,y2+y_size/4,bcol);
         m_bmp[img].DrawLine(x2,y2-y_size/4,x2-x_size/2,y2+y_size/4,bcol);

         m_bmp[img].DrawLine(x1,y1+y_size/4,x1,y2-y_size/4,bcol);
         m_bmp[img].DrawLine(x2,y1+y_size/4,x2,y2-y_size/4,bcol);
        }
     }
  }
//+------------------------------------------------------------------+
//|      ( m_bmp)                 |
//+------------------------------------------------------------------+
void CSOM::Render()
  {
   int ind=0;
   for(int i=0; i<m_xcells; i++)
     {
      for(int j=0; j<m_ycells; j++)
        {
         //---        
         for(int k=0; k<m_dimension; k++)
           {
            //m_som_nodes[ind].GetCoordinates(x1,y1,x2,y2);
            double range=(m_max_values[k]-m_min_values[k]);
            char avg=char(255*(m_som_nodes[ind].GetWeight(k)-m_min_values[k])/range);
            int col=int(GetPalColor(avg));
            RenderCell(k,col,ind,(j%2==0));
           }
         ind++;
        }
      //---        
      for(int m=0; m<m_dimension; m++)
        {
         for(int k=0; k<15; k++)
           {
            for(int l=0; l<m_xsize; l++)
              {
               uchar colind=uchar(MathRound(l*255.0/m_xsize));
               int col=int(GetPalColor(colind));
               m_bmp[m].DrawDot(l,k,col);
              }
           }
         //---         
         m_bmp[m].TypeText(0,0,DoubleToString(m_min_values[m],2),clrBlack);
         m_bmp[m].TypeText(m_xsize-25,0,DoubleToString(m_max_values[m],2),clrBlack);

         if(ColorScheme==2)
            m_bmp[m].TypeText(m_xsize/2-3*StringLen(m_som_titles[m]),0,m_som_titles[m],clrWhite);
         else
            m_bmp[m].TypeText(m_xsize/2-3*StringLen(m_som_titles[m]),0,m_som_titles[m],clrBlack);
        }
     }
   return;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM::NetDeinit()
  {
//---      
   ChartSetInteger(0,CHART_SHOW_GRID,m_gridmode);
//---   SOM  
   ObjectDelete(0,"SOM");
  }
//+------------------------------------------------------------------+
//|    ,  delimiter                |
//+------------------------------------------------------------------+
void CSOM::ParseString(string value,string delimiter,string  &mystrings[])
  {
   int indx;
   int dlen=StringLen(delimiter);
   string txt=value+delimiter;
   int strind=0;
   while(StringLen(txt)>0)
     {
      strind++;
      indx=StringFind(txt,delimiter);
      string tempstr=StringSubstr(txt,0,indx);
      txt=StringSubstr(txt,indx+dlen,-1);
      ArrayResize(mystrings,strind,50);
      mystrings[strind-1]=tempstr;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CSOM::ReadCSVData(int handle)
  {
   string stringsarr[];
   string s;

   int strings_total=0;
   while(!FileIsEnding(handle))
     {
      s=FileReadString(handle);
      if(StringLen(s)>0)
        {
         strings_total++;

         if(strings_total>1)
           {
            //Print("string ",IntegerToString(strings_total),": ",s);
            ParseString(s,";",stringsarr);
            double datavector[];
            int dimension=ArraySize(stringsarr)-1;
            ArrayResize(datavector,dimension);

            for(int i=0; i<dimension; i++)
              {
               datavector[i]=StringToDouble(stringsarr[i]);
               //Print(i," ",stringsarr[i]," data=",datavector[i]);
              }
            //--    
            AddVectorToTrainingSet(datavector,stringsarr[ArraySize(stringsarr)-1]);
           }
         else
           {
            ParseString(s,";",stringsarr);
            Print("HEADER: "+IntegerToString(ArraySize(stringsarr)));

            //---  ,  10000 
            //---  CellsX*CellsY ,   ImageXSize x ImageYSize
            int dimension=ArraySize(stringsarr)-1;
            InitParameters(dimension,10000,CellsX,CellsY,ImageXSize,ImageYSize);

            //---   
            ArrayResize(m_som_titles,ArraySize(stringsarr));
            for(int i=0; i<ArraySize(stringsarr); i++)
              {
               //stringsarr[i]=IntegerToString(i+1)+"_"+stringsarr[i];
               //Print(IntegerToString(i),": ",stringsarr[i]);
               m_som_titles[i]=stringsarr[i];
              }
           }
        }
     }

  }
//+------------------------------------------------------------------+
//|       filename          |
//+------------------------------------------------------------------+
bool CSOM::LoadTrainDataFromFile(string filename)
  {
//---
   ResetLastError();
   int filehandle=FileOpen(filename,FILE_READ|FILE_ANSI);
   if(filehandle!=INVALID_HANDLE)
     {
      Print("FileOpen OK");
      ReadCSVData(filehandle);
      FileClose(filehandle);
      return(true);
     }
   else
     {
      Print("   ",filename," : ",GetLastError());
      return(false);
     }
  }

CSOM KohonenMap;
//+------------------------------------------------------------------+
//| Expert initialization function                                   |
//+------------------------------------------------------------------+
int OnInit()
  {
   MathSrand(200);

//---     
   if(!KohonenMap.LoadTrainDataFromFile(DataFileName))
     {
      Print("    ");
      return(1);
     }
//---  
   KohonenMap.Train();
//---    
   KohonenMap.Render();
//---          
   KohonenMap.ShowTrainPatterns();
//---    
   KohonenMap.ShowBMP(false);
   return(0);
  }
//+------------------------------------------------------------------+
//| Expert deinitialization function                                 |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
   KohonenMap.NetDeinit();
  }
//+------------------------------------------------------------------+
//| Expert tick function                                             |
//+------------------------------------------------------------------+
void OnTick()
  {
//---
  }
//+------------------------------------------------------------------+
